package ru.dbotthepony.kstarbound.io.json

import com.google.common.collect.ImmutableMap
import com.google.common.collect.ImmutableSet
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import org.apache.logging.log4j.LogManager
import kotlin.reflect.KClass
import kotlin.reflect.KMutableProperty1
import kotlin.reflect.full.isSubclassOf

@Suppress("FunctionName")
fun <T : Any> BuilderAdapter(factory: () -> T, vararg fields: KMutableProperty1<T, *>): BuilderAdapter<T> {
	val builder = BuilderAdapter.Builder(factory)

	for (field in fields) {
		builder.auto(field)
	}

	return builder.build()
}

/**
 * [TypeAdapter] для классов, которые создаются "без всего", а после наполняются данными (паттерн builder).
 */
class BuilderAdapter<T : Any> private constructor(
	/**
	 * Завод по созданию объектов типа [T]
	 */
	val factory: () -> T,

	/**
	 * Свойства объекта [T], которые можно выставлять
	 */
	val properties: ImmutableMap<String, WrappedProperty<T, Any?>>,

	/**
	 * Ключи, которые необходимо игнорировать при чтении JSON
	 */
	val ignoreKeys: ImmutableSet<String>,
) : TypeAdapter<T>() {
	private val loggedMisses = ObjectOpenHashSet<String>()

	override fun write(writer: JsonWriter, value: T) {
		TODO("Not yet implemented")
	}

	override fun read(reader: JsonReader): T {
		val missing = ObjectArraySet<WrappedProperty<T, Any?>>()
		missing.addAll(properties.values)

		reader.beginObject()

		val instance = factory.invoke()

		while (reader.hasNext()) {
			val name = reader.nextName()

			if (ignoreKeys.contains(name)) {
				reader.skipValue()
				continue
			}

			val property = properties[name]

			if (property != null) {
				try {
					val peek = reader.peek()

					if (!property.returnType.isMarkedNullable && peek == JsonToken.NULL) {
						throw NullPointerException("Property ${property.property.name} of ${instance::class.qualifiedName} does not accept nulls")
					} else if (peek == JsonToken.NULL) {
						property.property.set(instance, null)
						reader.nextNull()
					} else {
						val readValue = property.adapter.read(reader)
						property.property.set(instance, readValue)
						check(missing.remove(property))
					}
				} catch(err: Throwable) {
					throw JsonSyntaxException("Reading property ${property.property.name} of ${instance::class.qualifiedName} near ${reader.path}", err)
				}
			} else {
				if (!loggedMisses.contains(name)) {
					LOGGER.warn("{} has no property for storing {}", instance::class.qualifiedName, name)
					loggedMisses.add(name)
				}

				reader.skipValue()
			}
		}

		reader.endObject()

		for (property in missing) {
			if (property.mustBePresent == null) {
				// null - проверяем, есть ли делегат
			}
		}

		return instance
	}

	data class WrappedProperty<T, V : Any?>(
		val property: KMutableProperty1<T, V>,
		val adapter: TypeAdapter<V>,
		val mustBePresent: Boolean?,
	) {
		// кеш
		val returnType = property.returnType
	}

	class PropertyConfigurator<T, V : Any?>(
		val property: KMutableProperty1<T, V>,
		val adapter: TypeAdapter<V>,
	) {
		/**
		 * Обязана ли присутствовать эта переменная внутри JSON структуры.
		 */
		var mustBePresent: Boolean? = null
	}

	class Builder<T : Any>(val factory: () -> T, vararg fields: KMutableProperty1<T, *>) {
		private val properties = ArrayList<WrappedProperty<T, *>>()
		private val ignoreKeys = ObjectArraySet<String>()

		init {
			for (field in fields)
				auto(field)
		}

		fun <V> add(property: KMutableProperty1<T, V>, adapter: TypeAdapter<V>, configurator: PropertyConfigurator<T, V>.() -> Unit = {}): Builder<T> {
			if (properties.any { it.property == property }) {
				throw IllegalArgumentException("Property $property is defined twice")
			}

			ignoreKeys.remove(property.name)

			val config = PropertyConfigurator(property, adapter)
			configurator.invoke(config)

			properties.add(WrappedProperty(
				property,
				adapter,
				mustBePresent = config.mustBePresent
			))

			return this
		}

		fun <V> auto(property: KMutableProperty1<T, V>, configurator: PropertyConfigurator<T, V>.() -> Unit = {}): Builder<T> {
			val returnType = property.returnType
			val classifier = returnType.classifier as? KClass<*> ?: throw ClassCastException("Unable to cast ${returnType.classifier} to KClass of property ${property.name}!")

			if (classifier.isSubclassOf(List::class)) {
				throw IllegalArgumentException("${property.name} is a List, please use autoList() method instead")
			}

			if (classifier.isSubclassOf(Map::class)) {
				throw IllegalArgumentException("${property.name} is a Map, please use autoMap() method instead")
			}

			@Suppress("unchecked_cast")
			return add(property, LazyTypeProvider(classifier.java) as TypeAdapter<V>, configurator)
		}

		fun ignoreKey(name: String): Builder<T> {
			if (properties.any { it.property.name == name }) {
				throw IllegalArgumentException("Can not ignore key $name because we have property with this name!")
			}

			ignoreKeys.add(name)
			return this
		}

		@Suppress("unchecked_cast")
		fun build(): BuilderAdapter<T> {
			val map = ImmutableMap.Builder<String, WrappedProperty<T, Any?>>()

			for (property in properties)
				map.put(property.property.name, property as WrappedProperty<T, Any?>)

			return BuilderAdapter(factory, map.build(), ImmutableSet.copyOf(ignoreKeys))
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger(ConfigurableTypeAdapter::class.java)
	}
}
